"""
symbolite.impl.libnumpy.lang
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

NumPy-backed implementations for language primitives.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from ..libstd.lang import *  # noqa: F403
